import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:build/build.dart';
import 'package:code_builder/code_builder.dart';
import 'package:dart_style/dart_style.dart';
import 'package:knt_isolate/knt_isolate.dart' as kntIsolate;
import 'package:source_gen/source_gen.dart';

import 'utils/helper.dart';
import 'utils/recase.dart';

Builder generatorFactoryBuilder(BuilderOptions options) =>
    SharedPartBuilder([KntIsolateGenerator()], 'knt_isolate');

class KntIsolateGenerator extends GeneratorForAnnotation<kntIsolate.KntIsolate>
    with _Constant {
  static const _frontendPostfix = 'Frontend';
  static const _backendPostfix = 'Backend';
  static const _eventPostfix = 'Event';

  // Builder
  final dartFormatter = DartFormatter();
  final dartEmitter = DartEmitter();

  // Details of assignable annotation class
  late String originClassName;
  late String className;
  late String frontendInit;
  late String backendInit;
  late String makeInstance;
  final List<String> interactingEvents = [];
  final workerIsolateMethods = <MethodElement>[];

  @override
  String generateForAnnotatedElement(
      Element element, ConstantReader annotation, BuildStep buildStep) {
    if (element is! ClassElement) {
      final name = element.displayName;
      throw InvalidGenerationSourceError(
        'Generator cannot target `$name`.',
        todo: 'Remove the [KntIsolate] annotation from `$name`.',
      );
    }
    return _implementClass(element, annotation);
  }

  String _implementClass(ClassElement element, ConstantReader? annotation) {
    // Get details of assignable annotation class
    final customClassName = annotation?.peek('serviceName')?.stringValue ?? '';
    originClassName = element.name;
    className = customClassName.isNotEmpty ? customClassName : originClassName;
    frontendInit = annotation?.peek('frontendInit')?.revive().accessor ?? '';
    backendInit = annotation?.peek('backendInit')?.revive().accessor ?? '';
    makeInstance = annotation?.peek('makeInstance')?.revive().accessor ?? '';
    interactingEvents.addAll(annotation
            ?.peek('interactingEvents')
            ?.listValue
            .map((e) => e.toStringValue() ?? '')
            .where((e) => e.isNotEmpty) ??
        []);
    workerIsolateMethods.addAll(element.methods);
    // Create Classes & Enum
    final enumBuilder = _buildEnum(element);
    final frontendClassBuilder = _buildFrontendClass(element);
    final backendClassBuilder = _buildBackendClass(element);
    // Generate code
    return [
      frontendClassBuilder,
      backendClassBuilder,
      _buildSpawnBackend,
      enumBuilder,
    ].join('\n\n');
  }
}

extension _BuildEnum on KntIsolateGenerator {
  String _buildEnum(ClassElement element) {
    if (workerIsolateMethods.isEmpty) {
      return '';
    }
    final values = workerIsolateMethods
        .map((method) => EnumValue((b) => b..name = method.generatedMethodName))
        .toList();
    values.addAll(
        interactingEvents.map((e) => EnumValue((b) => b..name = e.camelCase)));
    final enumBuilder = Enum((b) => b
      ..name = _enumName
      ..values.addAll([EnumValue((b) => b..name = 'init'), ...values]));
    return dartFormatter.format('${enumBuilder.accept(dartEmitter)}');
  }

  String get _enumName => '_\$$className${KntIsolateGenerator._eventPostfix}';
}

extension _BuildFrontend on KntIsolateGenerator {
  String _buildFrontendClass(ClassElement element) {
    final classBuilder = Class((c) {
      c
        ..name = _frontendName
        ..mixins.add(refer('Frontend'))
        ..methods.addAll([
          _initActionMethodForFrontend,
          _initMethodForFrontend,
          ...workerIsolateMethods.map(_parseMethodForFrontend).toList(),
          _sendMethod,
          _sendCouldNullMethod,
        ]);
    });
    return dartFormatter.format('${classBuilder.accept(dartEmitter)}');
  }

  Method _parseMethodForFrontend(MethodElement element) {
    final returnType = displayString(element.type.returnType);
    return Method((mm) {
      mm
        ..lambda = true
        ..name = element.generatedMethodName
        ..types.addAll(element.typeParameters.map((e) => refer(e.name)))
        ..returns = refer('Future<$returnType>');

      // Required parameters
      mm.requiredParameters.addAll(element.parameters
          .where((it) => it.isRequiredPositional)
          .map((it) => Parameter((p) => p
            ..type = refer(displayString(it.type))
            ..name = it.name
            ..named = it.isNamed)));

      // Optional positional or named parameters
      mm.optionalParameters.addAll(element.parameters
          .where((it) => it.isOptional || it.isRequiredNamed)
          .map((it) => Parameter((p) => p
            ..required = (it.isNamed &&
                it.type.nullabilitySuffix == NullabilitySuffix.none &&
                !it.hasDefaultValue)
            ..type = refer(displayString(it.type))
            ..name = it.name
            ..named = it.isNamed
            ..defaultTo = it.defaultValueCode == null
                ? null
                : Code(it.defaultValueCode!))));

      // Body
      final sendMethod =
          returnType.contains('void') ? _Constant.sendCouldNull : _Constant.send;
      final eventName = '$_enumName.${element.generatedMethodName}';
      final dataMap = {
        for (final param in element.parameters) '\'${param.name}\'': param.name,
      };
      final data = dataMap.isEmpty ? '' : ', KntIsolateUnit(data: $dataMap)';
      mm.body = Code('$sendMethod($eventName$data)');
    });
  }

  Method get _initActionMethodForFrontend => Method((mm) {
        mm
          ..name = 'initActions'
          ..annotations.add(CodeExpression(Code('override')))
          ..returns = refer('void');
        // Body
        mm.body = Code(interactingEvents
            .map((e) => e.camelCase)
            .map((e) => 'whenEventCome($_enumName.$e).run(_$e);')
            .join('\n'));
      });

  Method get _initMethodForFrontend => Method((mm) {
        mm
          ..name = 'init'
          ..modifier = MethodModifier.async
          ..returns = refer('Future<void>');
        mm.body = Code('''
        final data = await $frontendInit();
        await initBackend(initializer: $_spawnBackendName);
        await run(event: $_enumName.init, data: data);
        ''');
      });

  Method get _sendMethod => Method((mm) {
        mm
          ..docs.addAll([
            '/// Handle communication between Frontend & Backend in case Response is non-null type'
          ])
          ..name = _Constant.send
          ..types.addAll(_sendMethodGenericType)
          ..modifier = MethodModifier.async
          ..returns = refer('Future<Res>');
        mm.requiredParameters.add(Parameter((p) {
          p
            ..name = 'request'
            ..type = refer('Event')
            ..named = false;
        }));
        mm.optionalParameters.add(Parameter((p) {
          p
            ..name = 'data'
            ..type = refer('Req?')
            ..named = false;
        }));
        mm.body = Code('''
        final Maybe<Res> maybe = await run(event: request, data: data);
        return maybe.value;
        ''');
      });

  Method get _sendCouldNullMethod => Method((mm) {
        mm
          ..docs.add(
              '/// Handle communication between Frontend & Backend in case Response is void or nullable type')
          ..name = _Constant.sendCouldNull
          ..types.addAll(_sendMethodGenericType)
          ..modifier = MethodModifier.async
          ..returns = refer('Future<Res?>');
        mm.requiredParameters.add(Parameter((p) {
          p
            ..name = 'request'
            ..type = refer('Event')
            ..named = false;
        }));
        mm.optionalParameters.add(Parameter((p) {
          p
            ..name = 'data'
            ..type = refer('Req?')
            ..named = false;
        }));
        mm.body = Code('''
        final Maybe<Res> maybe = await run(event: request, data: data);
        return maybe.hasValue ? maybe.value : null;
        ''');
      });

  String get _frontendName =>
      '$className${KntIsolateGenerator._frontendPostfix}';

  List<Reference> get _sendMethodGenericType => [
        'Event',
        'Req extends Object?',
        'Res extends Object?'
      ].map((e) => refer(e)).toList();
}

extension _BuildBackend on KntIsolateGenerator {
  String _buildBackendClass(ClassElement element) {
    final classBuilder = Class((c) {
      c
        ..name = _backendClassName
        ..extend = refer('Backend')
        ..fields.add(Field((f) {
          f
            ..late = true
            ..type = refer(originClassName)
            ..name = _handlerVariable;
        }))
        ..constructors.add(Constructor((c) {
          c
            ..lambda = true
            ..requiredParameters.add(Parameter((p) {
              p
                ..name = 'argument'
                ..type = refer('BackendArgument<void>');
            }))
            ..initializers.add(Code('super(argument: argument)'));
        }))
        ..methods.addAll([
          _initActionMethodForBackend,
          _initMethodForBackend,
          ...workerIsolateMethods
              .where((method) {
                final annotation =
                    method.annotationOf(type: kntIsolate.KntIsolateEvent);
                if (annotation != null) {
                  // We will not generate code for ignored event
                  final ignored =
                      annotation.peek('ignored')?.boolValue ?? false;
                  return !ignored;
                }
                return true;
              })
              .map(_parseMethodForBackend)
              .toList(),
          _sendNonProtectedMethod,
        ]);
    });
    return dartFormatter.format('${classBuilder.accept(dartEmitter)}');
  }

  Method _parseMethodForBackend(MethodElement element) {
    return Method((mm) {
      mm
        ..lambda = true
        ..name = element.backendName
        ..types.addAll(element.typeParameters.map((e) => refer(e.name)))
        ..returns = refer(displayString(element.type.returnType));
      mm.optionalParameters.addAll(_commonParams);
      // Body
      final bodyPrefix = '$_handlerVariable.${element.name}(';
      final bodyInfix = element.parameters.map((p) {
        final paramName = p.name;
        final paramPresent = p.isNamed ? '$paramName: ' : '';
        return '${paramPresent}data.data[\'$paramName\'] as ${p.type.toString()}';
      }).join(',\n');
      final bodyPostfix = ')';
      mm.body = Code('$bodyPrefix$bodyInfix$bodyPostfix');
    });
  }

  Method get _initActionMethodForBackend => Method((mm) {
        mm
          ..name = 'initActions'
          ..annotations.add(CodeExpression(Code('override')))
          ..returns = refer('void');
        // Body
        final events = ['whenEventCome($_enumName.init).run(_init);'];
        events.addAll(workerIsolateMethods.map((method) =>
            'whenEventCome($_enumName.${method.generatedMethodName}).run(${method.backendName});'));
        mm.body = Code(events.join('\n'));
      });

  Method get _initMethodForBackend => Method((mm) {
        mm
          ..name = '_init'
          ..modifier = MethodModifier.async
          ..optionalParameters.addAll(_commonParams)
          ..returns = refer('Future<void>');
        // Body
        mm.body = Code('''
        await $backendInit(data);
        _plantHandler = $makeInstance();
        ''');
      });

  Method get _sendNonProtectedMethod => Method((mm) {
        mm
          ..lambda = true
          ..docs.add(
              '/// A wrapper for calling the protected method [Backend.send]')
          ..name = '_sendNonProtected'
          ..types.addAll(['Event', 'Data'].map((e) => refer(e)))
          ..returns = refer('Future<void>');

        // Optional positional or named parameters
        mm.optionalParameters.addAll([
          Parameter((p) => p
            ..required = true
            ..named = true
            ..type = refer('Event')
            ..name = 'event'),
          Parameter((p) => p
            ..named = true
            ..type = refer('Data?')
            ..name = 'data'),
          Parameter((p) => p
            ..named = true
            ..type = refer('bool')
            ..name = 'forceUpdate'
            ..defaultTo = Code('false')),
          Parameter((p) => p
            ..named = true
            ..type = refer('bool')
            ..name = 'sendDirectly'
            ..defaultTo = Code('false')),
        ]);

        // Body
        mm.body = Code('''
        send(
          event: event,
          data: data,
          forceUpdate: forceUpdate,
          sendDirectly: sendDirectly,
        )
        ''');
      });

  String get _buildSpawnBackend {
    return '''
    $_backendClassName $_spawnBackendName(BackendArgument<void> argument) =>
        $_backendClassName(argument);
    ''';
  }

  String get _spawnBackendName => '_spawn$_backendName';

  String get _backendClassName => '_$_backendName';

  String get _backendName => '$className${KntIsolateGenerator._backendPostfix}';

  String get _handlerVariable => '_${originClassName.camelCase}';

  List<Parameter> get _commonParams => [
        Parameter((p) {
          p
            ..named = true
            ..required = true
            ..name = 'event'
            ..type = refer(_enumName);
        }),
        Parameter((p) {
          p
            ..named = true
            ..required = true
            ..name = 'data'
            ..type = refer('KntIsolateUnit');
        }),
      ];
}

mixin _Constant {
  static const send = '_send';
  static const sendCouldNull = '_sendCouldNull';
}

extension CustomNameExt on MethodElement {
  String get generatedMethodName {
    var customName = '';
    final annotation = annotationOf(type: kntIsolate.KntIsolateEvent);
    if (annotation != null) {
      // We will not generate code for ignored event
      customName = annotation.peek('name')?.stringValue ?? '';
    }
    return customName.isNotEmpty ? customName : name;
  }
}

extension BackendMethodElement on MethodElement {
  String get backendName => '_handle${generatedMethodName.pascalCase}';
}
