import 'dart:ffi';

import 'package:analyzer/dart/constant/value.dart';
import 'package:analyzer/dart/element/element.dart';
import 'package:analyzer/dart/element/nullability_suffix.dart';
import 'package:analyzer/dart/element/type.dart';
import 'package:built_collection/built_collection.dart';
import 'package:code_builder/code_builder.dart';
import 'package:source_gen/source_gen.dart';

String displayString(dynamic e, {bool withNullability = false}) {
  try {
    return e.getDisplayString(withNullability: withNullability);
  } catch (error) {
    if (error is TypeError) {
      return e.getDisplayString();
    } else {
      rethrow;
    }
  }
}

/// Returns `$revived($args $kwargs)`, this won't have ending semi-colon (`;`).
/// [object] must not be null.
/// [object] is assumed to be a constant.
String revivedLiteral(
  Object object, {
  DartEmitter? dartEmitter,
}) {
  dartEmitter ??= DartEmitter();

  ArgumentError.checkNotNull(object, 'object');

  Revivable? revived;
  if (object is Revivable) {
    revived = object;
  }
  if (object is DartObject) {
    revived = ConstantReader(object).revive();
  }
  if (object is ConstantReader) {
    revived = object.revive();
  }
  if (revived == null) {
    throw ArgumentError.value(object, 'object',
        'Only `Revivable`, `DartObject`, `ConstantReader` are supported values');
  }

  String instantiation = '';
  final location = revived.source.toString().split('#');

  /// If this is a class instantiation then `location[1]` will be populated
  /// with the class name
  if (location.length > 1) {
    instantiation = location[1] +
        (revived.accessor.isNotEmpty ? '.${revived.accessor}' : '');
  } else {
    /// Getters, Setters, Methods can't be declared as constants so this
    /// literal must either be a top-level constant or a static constant and
    /// can be directly accessed by `revived.accessor`
    return revived.accessor;
  }

  final args = StringBuffer();
  final kwargs = StringBuffer();
  Spec objectToSpec(DartObject? object) {
    if (object == null) return literalNull;
    final constant = ConstantReader(object);
    if (constant.isNull) {
      return literalNull;
    }

    if (constant.isBool) {
      return literal(constant.boolValue);
    }

    if (constant.isDouble) {
      return literal(constant.doubleValue);
    }

    if (constant.isInt) {
      return literal(constant.intValue);
    }

    if (constant.isString) {
      return literal(constant.stringValue);
    }

    if (constant.isList) {
      return literalList(constant.listValue.map(objectToSpec));
      // return literal(constant.listValue);
    }

    if (constant.isMap) {
      return literalMap(Map.fromIterables(
          constant.mapValue.keys.map(objectToSpec),
          constant.mapValue.values.map(objectToSpec)));
      // return literal(constant.mapValue);
    }

    if (constant.isSymbol) {
      return Code('Symbol(${constant.symbolValue.toString()})');
      // return literal(constant.symbolValue);
    }

    if (constant.isNull) {
      return literalNull;
    }

    if (constant.isType) {
      return refer(displayString(constant.typeValue));
    }

    if (constant.isLiteral) {
      return literal(constant.literalValue);
    }

    /// Perhaps an object instantiation?
    /// In that case, try initializing it and remove `const` to reduce noise
    final revived = revivedLiteral(constant.revive(), dartEmitter: dartEmitter)
        .replaceFirst('const ', '');
    return Code(revived);
  }

  for (var arg in revived.positionalArguments) {
    final literalValue = objectToSpec(arg);

    args.write('${literalValue.accept(dartEmitter)},');
  }

  for (var arg in revived.namedArguments.keys) {
    final literalValue = objectToSpec(revived.namedArguments[arg]!);

    kwargs.write('$arg:${literalValue.accept(dartEmitter)},');
  }

  return '$instantiation($args $kwargs)';
}

TypeChecker typeChecker(Type type) => TypeChecker.fromRuntime(type);

extension DartTypeStreamAnnotation on DartType {
  bool get isDartAsyncStream {
    final element = this.element == null ? null : this.element as ClassElement;
    if (element == null) {
      return false;
    }
    return element.name == 'Stream' && element.library.isDartAsync;
  }
}

extension DartTypeExt on DartType {
  bool get isBasicType {
    return typeChecker(String).isExactlyType(this) ||
        typeChecker(bool).isExactlyType(this) ||
        typeChecker(int).isExactlyType(this) ||
        typeChecker(double).isExactlyType(this) ||
        typeChecker(num).isExactlyType(this) ||
        typeChecker(Double).isExactlyType(this) ||
        typeChecker(Float).isExactlyType(this);
  }

  bool get isBasicInnerType => genericOf?.isBasicType ?? false;

  bool get isNullable => this.nullabilitySuffix == NullabilitySuffix.question;

  List<DartType>? get genericListOf => this is ParameterizedType &&
          (this as ParameterizedType).typeArguments.isNotEmpty
      ? (this as ParameterizedType).typeArguments
      : null;

  DartType? get genericOf =>
      this is InterfaceType && (this as InterfaceType).typeArguments.isNotEmpty
          ? (this as InterfaceType).typeArguments.first
          : null;

  DartType? get responseType => this.genericOf;

  /// get types for `Map<String, List<User>>`, `A<B,C,D>`
  List<DartType>? get responseInnerTypes => this.genericListOf;

  DartType? get responseInnerType {
    final generic = this.genericOf;
    if (generic == null ||
        typeChecker(Map).isExactlyType(this) ||
        typeChecker(BuiltMap).isExactlyType(this)) {
      return this;
    }

    if (generic.isDynamic) {
      return null;
    }

    if (typeChecker(List).isExactlyType(this) ||
        typeChecker(BuiltList).isExactlyType(this)) {
      return generic;
    }

    return generic.responseInnerType;
  }
}

extension ReferenceExt on Reference {
  Reference asNoNull() => refer('${this.symbol}!');

  Reference asNoNullIf({required bool returnNullable}) =>
      returnNullable ? this : this.asNoNull();

  Expression propertyIf({
    required bool thisNullable,
    required String name,
  }) =>
      thisNullable
          ? this.nullSafeProperty(name)
          : this.asNoNull().property(name);

  Expression conditionalIsNullIf({
    required bool thisNullable,
    required Expression whenFalse,
  }) =>
      thisNullable
          ? this.equalTo(literalNull).conditional(literalNull, whenFalse)
          : whenFalse;
}

extension MethodElementExt on MethodElement {
  ConstantReader? annotationOf({required Type type}) {
    final annotation =
        typeChecker(type).firstAnnotationOf(this, throwOnUnresolved: false);
    if (annotation != null) {
      return ConstantReader(annotation);
    }
    return null;
  }
}

extension IterableExtension<T> on Iterable<T> {
  T? firstWhereOrNull(bool Function(T) test) {
    for (T element in this) {
      if (test(element)) return element;
    }
    return null;
  }
}
